@@ -9,11 +9,14 @@ package mod
 import (
 	"context"
 	"fmt"
+	"runtime"
 	"sort"
 	"strings"
+	"sync"
 
 	"golang.org/x/mod/modfile"
 	"golang.org/x/mod/semver"
+	"golang.org/x/sync/errgroup"
 	"golang.org/x/tools/gopls/internal/govulncheck"
 	"golang.org/x/tools/gopls/internal/lsp/command"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
@@ -58,24 +61,36 @@ func VulnerabilityDiagnostics(ctx context.Context, snapshot source.Snapshot) (ma
 }
 
 func collectDiagnostics(ctx context.Context, snapshot source.Snapshot, diagFn func(context.Context, source.Snapshot, source.FileHandle) ([]*source.Diagnostic, error)) (map[span.URI][]*source.Diagnostic, error) {
+
+	g, ctx := errgroup.WithContext(ctx)
+	cpulimit := runtime.GOMAXPROCS(0)
+	g.SetLimit(cpulimit)
+
+	var mu sync.Mutex
 	reports := make(map[span.URI][]*source.Diagnostic)
+
 	for _, uri := range snapshot.ModFiles() {
-		fh, err := snapshot.ReadFile(ctx, uri)
-		if err != nil {
-			return nil, err
-		}
-		reports[fh.URI()] = []*source.Diagnostic{}
-		diagnostics, err := diagFn(ctx, snapshot, fh)
-		if err != nil {
-			return nil, err
-		}
-		for _, d := range diagnostics {
-			fh, err := snapshot.ReadFile(ctx, d.URI)
+		uri := uri
+		g.Go(func() error {
+			fh, err := snapshot.ReadFile(ctx, uri)
 			if err != nil {
-				return nil, err
+				return err
 			}
-			reports[fh.URI()] = append(reports[fh.URI()], d)
-		}
+			diagnostics, err := diagFn(ctx, snapshot, fh)
+			if err != nil {
+				return err
+			}
+			for _, d := range diagnostics {
+				mu.Lock()
+				reports[d.URI] = append(reports[fh.URI()], d)
+				mu.Unlock()
+			}
+			return nil
+		})
+	}
+
+	if err := g.Wait(); err != nil {
+		return nil, err
 	}
 	return reports, nil
 }
@@ -316,7 +331,13 @@ func ModVulnerabilityDiagnostics(ctx context.Context, snapshot source.Snapshot,
 	// TODO(hyangah): place this diagnostic on the `go` directive or `toolchain` directive
 	// after https://go.dev/issue/57001.
 	const diagnoseStdLib = false
-	if diagnoseStdLib {
+
+	// If diagnosing the stdlib, add standard library vulnerability diagnostics
+	// on the module declaration.
+	//
+	// Only proceed if we have a valid module declaration on which to position
+	// the diagnostics.
+	if diagnoseStdLib && pm.File.Module != nil && pm.File.Module.Syntax != nil {
 		// Add standard library vulnerabilities.
 		stdlibVulns := vulnsByModule["stdlib"]
 		if len(stdlibVulns) == 0 {
@@ -516,7 +537,7 @@ func SelectUpgradeCodeActions(actions []protocol.CodeAction) []protocol.CodeActi
 	var chosenVersionedUpgrade string
 	var selected []protocol.CodeAction
 
-	seen := make(map[string]bool)
+	seenTitles := make(map[string]bool)
 
 	for _, action := range actions {
 		if strings.HasPrefix(action.Title, upgradeCodeActionPrefix) {
@@ -528,8 +549,8 @@ func SelectUpgradeCodeActions(actions []protocol.CodeAction) []protocol.CodeActi
 			}
 		} else if strings.HasPrefix(action.Title, "Reset govulncheck") {
 			resetAction = action
-		} else if !seen[action.Command.Title] {
-			seen[action.Command.Title] = true
+		} else if !seenTitles[action.Command.Title] {
+			seenTitles[action.Command.Title] = true
 			selected = append(selected, action)
 		}
 	}
